import requests
import json
import datetime
import webbrowser
import logging
import os

# from .callbackHandler import *
# Gets or creates a logger
logger = logging.getLogger(__name__)

# set log level
logger.setLevel(logging.DEBUG)
# define file handler and set formatter
log_fileName = 'LOG/{:%Y-%m-%d}.log'.format(datetime.datetime.now())
os.makedirs(os.path.dirname(log_fileName), exist_ok=True)
file_handler = logging.FileHandler(log_fileName, mode="a", encoding=None, )
formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(funcName)s  %(message)s')
file_handler.setFormatter(formatter)
# add file handler to logger
logger.addHandler(file_handler)


class ShurjoPay(object):
    _POST_DEFAULT_ADDRESS = "https://sandbox.shurjopayment.com"
    _TOKEN_END_POINT = "/api/get_token"
    _VERIFICATION_END_POINT = "/api/verification"
    _PAYMENT_STATUS_END_POINT = "/api/payment-status"
    _SANDBOX_USERNAME = "sp_sandbox"
    _SANDBOX_PASSWORD = "pyyk97hu&6u6"
    # _DEFAULT_RETURN_URL = "http://0.0.0.0:8000/en/sharedfarm/test_return_url"
    # _DEFAULT_CANCEL_URL = "http://0.0.0.0:8000/en/sharedfarm/test_return_url"
    _DEFAULT_RETURN_URL = "http://0.0.0.0:8008/return"
    _DEFAULT_CANCEL_URL = "http://0.0.0.0:8008/cancel"
    _DEFAULT_CLIENT_IP = "192.168.0.99"

    def __init__(self, prefix, currency, return_url=None, cancel_url=None, client_ip=None, username=None, password=None,
                 post_address=None, ):
        """

        :param prefix: Any string not more than 5 characters. It distinguishes the stores of a merchant.
        :param currency: ISO format,(only BDT and USD are allowed).
        :param return_url: Merchant should provide a GET Method return url to verify users initiated transaction status.
        :param cancel_url: Merchant should provide a cancel url to redirect the user if user cancels the transaction in midway.
        :param client_ip: User's ip
        :param username: Merchant Username provided by shurjopay.
        :param password: Merchant Password provided by shurjopay.
        :param post_address: Live shurjopay version 2 URL.
        """
        self.sandbox = False
        if post_address:
            self.root_address = post_address
            pass
        else:
            self.sandbox = True
            self.root_address = self._POST_DEFAULT_ADDRESS
            pass
        if client_ip:
            self.client_ip = client_ip
            pass
        else:
            self.client_ip = self._DEFAULT_CLIENT_IP
            pass
        if return_url:
            self.return_url = return_url
            pass
        else:
            self.return_url = self._DEFAULT_RETURN_URL
            pass
        if cancel_url:
            self.cancel_url = cancel_url
            pass
        else:
            self.cancel_url = self._DEFAULT_CANCEL_URL
            pass
        if username and password:

            self.username = username
            self.password = password
            pass
        else:
            self.username = self._SANDBOX_USERNAME
            self.password = self._SANDBOX_PASSWORD
            pass
        self.prefix = prefix
        self.currency = currency

        # declaration of token info
        self._token = None
        self._store_id = None
        self._execute_url = None
        self._expires_in = None
        self._token_type = None
        self._token_create_time = None
        self.set_sp_token()

        pass

    def set_sp_token(self):
        try:
            payload = {
                "username": self.username,
                "password": self.password
            }
            logger.info(f"url: {self.root_address + self._TOKEN_END_POINT}, payload:{payload}")
            response = requests.post(self.root_address + self._TOKEN_END_POINT, data=payload)
            # {'token': 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwczpcL1wvc2FuZGJveC5zaHVyam9wYXltZW50LmNvbVwvYXBpXC9sb2dpbiIsImlhdCI6MTYzNTE0NTIxMywiZXhwIjoxNjM1MTQ4ODEzLCJuYmYiOjE2MzUxNDUyMTMsImp0aSI6IlVZZ1FkRWtSS2ZnNEJybHoiLCJzdWIiOjEsInBydiI6IjgwNWYzOWVlZmNjNjhhZmQ5ODI1YjQxMjI3ZGFkMGEwNzZjNDk3OTMifQ.cPXzg0UqeMD_6hVxqF1oKnzkwaJeRIn1gaEejAPz100', 'store_id': 1, 'execute_url': 'https://sandbox.shurjopayment.com/api/secret-pay', 'token_type': 'Bearer', 'sp_code': '200', 'massage': 'Ok', 'TokenCreateTime': '2021-10-25 01:00:13pm', 'expires_in': 3600}

            response_json = response.json()
            # set token information
            logger.info(f"response: {response_json}")
            if int(response_json["sp_code"]) == 200:
                self._token = response_json["token"]
                self._store_id = response_json["store_id"]
                self._execute_url = response_json["execute_url"]
                self._expires_in = response_json["expires_in"]
                self._token_type = response_json["token_type"]
                self._token_create_time = datetime.datetime.strptime(response_json["TokenCreateTime"],
                                                                     "%Y-%m-%d %I:%M:%S%p")

                pass
            else:
                raise Exception(response_json)
            return True
        except Exception as e:
            logger.exception(e)
            pass

    def _check_token_validity(self):
        if self.sandbox:
            return True
        else:
            if (self._token_create_time + datetime.timedelta(milliseconds=self._expires_in)) > datetime.datetime.now():
                return True
            else:
                return False

    def get_token(self):
        return self._token

    def checkout_web(self, amount, order_id, customer_name, customer_phone, customer_address, customer_city):
        """

        :param amount:
        :param order_id:
        :param customer_name:
        :param customer_phone:
        :param customer_address:
        :param customer_city:
        :return:
        """
        try:

            _headers = {'content-type': 'application/json', 'Authorization': f'{self._token_type} {self._token}'}
            _payloads = {
                "token": self._token,
                "store_id": self._store_id,
                "prefix": self.prefix,
                "currency": self.currency,
                "return_url": self.return_url,
                "cancel_url": self.cancel_url,
                "amount": amount,
                "order_id": order_id,
                "customer_name": customer_name,
                "customer_phone": customer_phone,
                "customer_address": customer_address,
                "customer_city": customer_city,
                "client_ip": self.client_ip
            }
            if self._check_token_validity():
                logger.info(f'url: {self._execute_url}, header:{_headers}, payload: {_payloads}')
                response = requests.post(self._execute_url, headers=_headers, data=json.dumps(_payloads))
                response_json = response.json()
                # driver =
                # r_response = requests.get(response_json["checkout_url"])
                # print(r_response.history)
                logger.info(f"response: {response_json}")
                webbrowser.open(response_json["checkout_url"])
                # wait_for_request('', 8008, self._token)
            else:
                logger.warning(f"is_token_expired: {self._check_token_validity()}")
            pass
        except Exception as e:
            logger.exception(e)
        pass

    def checkout(self, amount, order_id, customer_name, customer_phone, customer_address, customer_city):
        """

        :param amount: Transaction Amount
        :param order_id: Provide a Unique ORDER ID
        :param customer_name: Customer Full Name
        :param customer_phone: Customer Active Contact Number
        :param customer_address: Customer Address
        :param customer_city: Customer City
        :return:
        """
        try:

            _headers = {'content-type': 'application/json', 'Authorization': f'{self._token_type} {self._token}'}
            _payloads = {
                "token": self._token,
                "store_id": self._store_id,
                "prefix": self.prefix,
                "currency": self.currency,
                "return_url": self.return_url,
                "cancel_url": self.cancel_url,
                "amount": amount,
                "order_id": order_id,
                "customer_name": customer_name,
                "customer_phone": customer_phone,
                "customer_address": customer_address,
                "customer_city": customer_city,
                "client_ip": self.client_ip
            }
            if self._check_token_validity():
                logger.info(f'url: {self._execute_url}, header:{_headers}, payload: {_payloads}')
                response = requests.post(self._execute_url, headers=_headers, data=json.dumps(_payloads))
                response_json = response.json()
                logger.info(f"response: {response_json}")
                return response_json
            else:
                logger.warning(f"is_token_expired: {self._check_token_validity()}")
                return {"message": "token expired!"}
        except Exception as e:
            logger.exception(e)
        pass

    def verifty(self, order_id):
        """

        :param order_id: For verification, merchants must use order_id that is sent from shurjoPay during call back.
        :return:
        """
        try:
            _headers = {'content-type': 'application/json', 'Authorization': f'{self._token_type} {self._token}'}
            _payloads = {
                "order_id": order_id,
            }
            if self._check_token_validity():
                logger.info(f'url: {self.root_address}, header:{_headers}, payload: {_payloads}')
                response = requests.post(self.root_address + self._VERIFICATION_END_POINT, headers=_headers,
                                         data=json.dumps(_payloads))
                response_json = response.json()
                logger.info(f"response: {response_json}")
                return response_json
            else:
                logger.warning(f"is_token_expired: {self._check_token_validity()}")
                return {"message": "token expired!"}
        except Exception as e:
            logger.exception(e)
        pass

    def check_status(self, order_id):
        """

        :param order_id:For transaction status, merchants must use order_id that is sent from shurjoPay during callback.
        :return:
        """
        try:
            _headers = {'content-type': 'application/json', 'Authorization': f'{self._token_type} {self._token}'}
            _payloads = {
                "order_id": order_id,
            }
            if self._check_token_validity():
                logger.info(f'url: {self.root_address}, header:{_headers}, payload: {_payloads}')
                response = requests.post(self.root_address + self._PAYMENT_STATUS_END_POINT, headers=_headers,
                                         data=json.dumps(_payloads))
                response_json = response.json()
                logger.info(f"response: {response_json}")
                return response_json
            else:
                logger.warning(f"is_token_expired: {self._check_token_validity()}")
                return {"message": "token expired!"}
        except Exception as e:
            logger.exception(e)
        pass



from collections import namedtuple
from contextlib import closing
from io import BytesIO
from json import dumps as json_encode
import os
import sys
import requests
import json
import datetime
import logging
import os
# Gets or creates a logger
logger = logging.getLogger(__name__)

# set log level
logger.setLevel(logging.DEBUG)
# define file handler and set formatter
log_fileName = 'LOGS/{:%Y-%m-%d}.log'.format(datetime.datetime.now())
os.makedirs(os.path.dirname(log_fileName), exist_ok=True)
file_handler = logging.FileHandler(log_fileName, mode="a", encoding=None,)
formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(funcName)s  %(message)s')
file_handler.setFormatter(formatter)
# add file handler to logger
logger.addHandler(file_handler)

if sys.version_info >= (3, 0):
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from socketserver import ThreadingMixIn
    from urllib.parse import parse_qs
else:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
    from SocketServer import ThreadingMixIn
    from urlparse import parse_qs

ResponseStatus = namedtuple("HTTPStatus",
                            ["code", "message"])

ResponseData = namedtuple("ResponseData",
                          ["status", "content_type", "data_stream"])

# Mapping the output format used in the client to the content type for the
# response
AUDIO_FORMATS = {"ogg_vorbis": "audio/ogg",
                 "mp3": "audio/mpeg",
                 "pcm": "audio/wave; codecs=1"}
CHUNK_SIZE = 1024
HTTP_STATUS = {"OK": ResponseStatus(code=200, message="OK"),
               "BAD_REQUEST": ResponseStatus(code=400, message="Bad request"),
               "NOT_FOUND": ResponseStatus(code=404, message="Not found"),
               "INTERNAL_SERVER_ERROR": ResponseStatus(code=500, message="Internal server error")}
PROTOCOL = "http"
RETURN_URL = "/return"
CANCEL_URL = "/cancel"


class HTTPStatusError(Exception):
    """Exception wrapping a value from http.server.HTTPStatus"""

    def __init__(self, status, description=None):
        """
        Constructs an error instance from a tuple of
        (code, message, description), see http.server.HTTPStatus
        """
        super(HTTPStatusError, self).__init__()
        self.code = status.code
        self.message = status.message
        self.explain = description


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """An HTTP Server that handle each request in a new thread"""
    daemon_threads = True


class ChunkedHTTPRequestHandler(BaseHTTPRequestHandler):
    """"HTTP 1.1 Chunked encoding request handler"""
    # Use HTTP 1.1 as 1.0 doesn't support chunked encoding
    protocol_version = "HTTP/1.1"

    verification_token = ''

    def query_get(self, queryData, key, default=""):
        """Helper for getting values from a pre-parsed query string"""
        return queryData.get(key, [default])[0]

    def do_HEAD(self):
        self.send_headers()

    def do_GET(self):
        """Handles GET requests"""

        # Extract values from the query string
        path, _, query_string = self.path.partition('?')
        query = parse_qs(query_string)

        response = None

        print(u"[START]: Received GET for %s with query: %s" % (path, query))

        try:
            # Handle the possible request paths
            if path == RETURN_URL:
                response = self.route_return(path, query)
            elif path == CANCEL_URL:
                response = self.route_cancel(path, query)
            else:
                response = self.route_not_found(path, query)

            self.send_headers(response.status, response.content_type)
            # self.stream_data(response.data_stream)
            logger.info(response)
            self._json(response.data_stream)

        except HTTPStatusError as err:
            # Respond with an error and log debug
            # information
            if sys.version_info >= (3, 0):
                self.send_error(err.code, err.message, err.explain)
            else:
                self.send_error(err.code, err.message)

            self.log_error(u"%s %s %s - [%d] %s", self.client_address[0],
                           self.command, self.path, err.code, err.explain)

        print("[END]")

    def route_not_found(self, path, query):
        """Handles routing for unexpected paths"""
        raise HTTPStatusError(HTTP_STATUS["NOT_FOUND"], "Page not found")

    def route_return(self, path, query):
        """Handles routing for the application's entry point'"""
        try:
            _POST_DEFAULT_ADDRESS = "https://sandbox.shurjopayment.com"
            _VERIFICATION_END_POINT = "/api/verification"
            # print('here!', query['order_id'][0])
            _headers = {'content-type': 'application/json', 'Authorization': f'Bearer {self.verification_token}'}
            _payloads = {
                "order_id": query['order_id'][0],
            }
            response = requests.post(_POST_DEFAULT_ADDRESS + _VERIFICATION_END_POINT, headers=_headers,
                                     data=json.dumps(_payloads))
            response_json = response.json()
            return ResponseData(status=HTTP_STATUS["OK"], content_type="application/json",
                                # Open a binary stream for reading the index
                                # HTML file
                                data_stream=response_json)
        except IOError as err:
            # Couldn't open the stream
            raise HTTPStatusError(HTTP_STATUS["INTERNAL_SERVER_ERROR"],
                                  str(err))

    def route_cancel(self, path, query):
        """Handles routing for the application's entry point'"""
        try:
            return ResponseData(status=HTTP_STATUS["OK"], content_type="application/json",
                                # Open a binary stream for reading the index
                                # HTML file
                                data_stream=open(os.path.join(sys.path[0],
                                                              path[1:]), "rb"))
        except IOError as err:
            # Couldn't open the stream
            raise HTTPStatusError(HTTP_STATUS["INTERNAL_SERVER_ERROR"],
                                  str(err))

    def send_headers(self, status, content_type):
        """Send out the group of headers for a successful request"""
        # Send HTTP headers

        self.send_response(status.code, status.message)
        self.send_header('Content-type', content_type)
        # self.send_header('Transfer-Encoding', 'chunked')
        # self.send_header('Connection', 'close')
        self.send_header('Access-Control-Allow-Origin', '*')
        self.end_headers()

    def _html(self, message):
        """This just generates an HTML document that includes `message`
        in the body. Override, or re-write this do do more interesting stuff.
        """
        content = f"<html><body><h1>{message}</h1></body></html>"
        return content.encode("utf8")  # NOTE: must return a bytes object!

    def _json(self, response):
        str = json_encode(response)
        self.wfile.write(str.encode("utf8"))

    def stream_data(self, stream):
        """Consumes a stream in chunks to produce the response's output'"""
        print("Streaming started...")

        if stream:
            # Note: Closing the stream is important as the service throttles on
            # the number of parallel connections. Here we are using
            # contextlib.closing to ensure the close method of the stream object
            # will be called automatically at the end of the with statement's
            # scope.
            # with closing(stream) as managed_stream:
            # Push out the stream's content in chunks
            while True:
                # data = managed_stream.read(CHUNK_SIZE)
                data = stream
                self.wfile.write(self._html(data))

                # If there's no more data to read, stop streaming
                if not data:
                    break

            # Ensure any buffered output has been transmitted and close the
            # stream
            self.wfile.flush()

            print("Streaming completed.")
        else:
            # The stream passed in is empty
            self.wfile.write(b"0\r\n\r\n")
            print("Nothing to stream.")


def wait_for_request(host, port, token):
    # Create and configure the HTTP server instance
    handler = ChunkedHTTPRequestHandler
    handler.verification_token = token
    server = ThreadedHTTPServer((host, port),
                                handler)
    print("Starting server, use <Ctrl-C> to stop...")
    print(u"Open {0}://{1}:{2} in a web browser.".format(PROTOCOL,
                                                         host,
                                                         port,
                                                         ))

    try:
        # Listen for requests indefinitely
        server.handle_request()

    except KeyboardInterrupt:
        # A request to terminate has been received, stop the server
        print("\nShutting down...")
        server.socket.close()
